/*
 * dl.c
 * dynamic runtime library loading routines
 */

#include <dlfcn.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "struct.h"
#include "log.h"
#include "dl.h"
#include "gc.h"

Module *module_list;
LD_Path *ld_path_list;

extern void broadcast_to_ops(char *fmt,...);
#define broadcast_to_opers broadcast_to_ops

/* void __init_mod_list()
   initializes module_list and ld_path_list - should be called at startup
   never call this from a user routine!
 */

void __init_mod_list()
{
     module_list = (Module *) gcAlloc(sizeof(Module));
     bzero(module_list, sizeof(Module));
     module_list->prev = NULL;
     module_list->next = NULL;
     ld_path_list = (LD_Path *) gcAlloc(sizeof(LD_Path));
     bzero(ld_path_list, sizeof(LD_Path));
     ld_path_list->prev = NULL;
     ld_path_list->next = NULL;
};

/* int add_ld_path(char *)
   adds a path entry to the list of paths to search for loadable modules. the paths
   are searched in order from the first one added to the last one.
   returns 0 on success, -1 on failure (always succeeds - failure not implemented..)
 */

int add_ld_path(char *path)
{
     LD_Path *path_ent, *list;

     path_ent = (LD_Path *) gcAlloc(sizeof(LD_Path));
     bzero(path_ent, sizeof(LD_Path));

     strncpy(path_ent->dl_path, path, 100);

     list = ld_path_list;
     while (list->next != NULL)
	  list = list->next;
     list->next = path_ent;
     path_ent->prev = list;
     return 0;
}

/* int load_module(char *)
   loads the specified module. the parameter is either the name of a module file or the full path to it. if
   if the full path is not provided, or the module is not found in the specified path, the directories added
   in add_ld_path() will be searched for the file.

   returns 0 on success, -1 on failure.
 */

int load_module(char *path)
{
     char *dl_error = NULL;
     void *dl_handle = NULL;
     Module_Info *(*mod_get_info) () = NULL;
     Functions *(*mod_get_funcs) () = NULL;
     EventFnList *(*mod_get_events) () = NULL;
     int (*mod_register_handler) (struct fd_list *) = NULL;
     Module_Info *mod_info_ptr = NULL;
     Functions *mod_funcs_ptr = NULL;
     EventFnList *event_fn_ptr = NULL;
     Module *mod_ptr = NULL, *list_ptr = NULL;

     log("opening module");
     fprintf(stderr, "module name: %s\n", path);
     dl_handle = dlopen(path, RTLD_LAZY);
     if (!dl_handle) {
	  LD_Path *list;

	  list = ld_path_list->next;
	  while (list != NULL) {
	       char p[255];

	       snprintf(p, 255, "%s/%s", list->dl_path, path);
	       dl_handle = dlopen(p, RTLD_LAZY);
	       if (!dl_handle) {
		    broadcast_to_opers("dlerror(): %s", dlerror());
		    list = list->next;
	       } else {
		    list = NULL;
	       }
	  }
	  if (!dl_handle) {
               broadcast_to_opers("dlerror(): %s", dlerror());
	       return -1;
	  }
     }
     log("loading symbols");
     mod_get_info = dlsym(dl_handle, "__module_get_info");
     if ((dl_error = dlerror()) != NULL) {
	  broadcast_to_opers("dlerror(): %s", dlerror());
	  dlclose(dl_handle);
	  return -1;
     }
     mod_get_funcs = dlsym(dl_handle, "__module_get_functions");
     if ((dl_error = dlerror()) != NULL) {
          broadcast_to_opers("dlerror(): %s", dlerror());
	  dlclose(dl_handle);
	  return -1;
     }
     mod_get_events = dlsym(dl_handle, "__module_get_events");
     /* no error check here - this one isn't essential to the functioning of the module */
     mod_register_handler = dlsym(dl_handle, "__module_register_handler");

     log("retreiving structures");
     mod_info_ptr = (*mod_get_info) ();
     mod_funcs_ptr = (*mod_get_funcs) ();
     if (mod_get_events)
	  event_fn_ptr = (*mod_get_events) ();

     if (mod_info_ptr == NULL || mod_funcs_ptr == NULL) {
	  dlclose(dl_handle);
	  broadcast_to_opers("%s: Could not load dynamic library %s!", __PRETTY_FUNCTION__, path);
	  broadcast_to_opers("%s: No %s avaliable!", __PRETTY_FUNCTION__, (!mod_info_ptr ? !mod_funcs_ptr ? 
			"mod_info_ptr or mod_funcs_ptr" : "mod_info_ptr" : "mod_funcs_ptr"));
	  return -1;
     }
     log("adding module to end of list");
     mod_ptr = (Module *) gcAlloc(sizeof(Module));
     if (mod_ptr == NULL) {
	  broadcast_to_opers("%s: Out of memory!\n", __PRETTY_FUNCTION__);
	  dlclose(dl_handle);
	  exit(1);
     }
     list_ptr = module_list;
     while (list_ptr->next != NULL)
	  list_ptr = list_ptr->next;

     bzero(mod_ptr, sizeof(Module));

     broadcast_to_opers("%s: internal module name: %s\n", __PRETTY_FUNCTION__, mod_info_ptr->module_name);
     broadcast_to_opers("%s: module description: %s\n", __PRETTY_FUNCTION__, mod_info_ptr->module_description);

     mod_ptr->prev = list_ptr;
     list_ptr->next = mod_ptr;
     mod_ptr->info = mod_info_ptr;
     mod_ptr->function_list = mod_funcs_ptr;
     mod_ptr->register_handler = mod_register_handler;
     mod_ptr->dl_handle = dl_handle;
     mod_ptr->other_funcs = event_fn_ptr;

     return 0;
};

/* int unload_module(char *)
   unloads the specified module, freeing memory allocated for the structure
   takes the name of the -module-, not of the module file!
   returns 0 on success, -1 on failure
 */

int unload_module(char *module_name)
{
     Module *list;

     list = module_list->next;

     while (list != NULL) {
	  if (!strcasecmp(list->info->module_name, module_name)) {
	       dlclose(list->dl_handle);
	       if (list->next != NULL) {
		    (list->prev)->next = list->next;
		    (list->next)->prev = list->prev;
	       } else {
		    (list->prev)->next = NULL;
	       }
	       gcDecRefcnt(list);
	       return 0;
	  }
	  list = list->next;
     }

     return -1;
}
